#ifndef C_P1CAMERA_COMMON_STRUCTS_H
#define C_P1CAMERA_COMMON_STRUCTS_H

/**
* @file P1CameraSdkCommonStructs.h
* @brief This file contains struct types that a used by higher layer APIs, like C++ and C#
*        Types defined here are include in the public headers for our C++ API layer.
*/

#include <stdint.h>
#include <string.h>  // For memcmp()
#include <stdexcept>
#include <string>
#include <vector>
#include <memory>
#include <ostream>

#include "C_P1CameraCommonDefs.h"	// For CameraId
#include "C_P1CameraCommonProperties.h"	// For PropertyId
#include "C_P1CameraCommonLinkage.h"
#include "C_P1CameraCommonEventTypes.h"

namespace P1
{
namespace CameraSdk
{
#pragma pack(push, 8)

	struct EXPORT_API Version
	{
		uint32_t major = 0;
		uint32_t minor = 0;
		uint32_t build = 0;
	};

	struct EXPORT_API ValueColorFloat
	{
		double alpha = 0.0;
		double red = 0.0;
		double green = 0.0;
		double blue = 0.0;

		bool operator==(ValueColorFloat const& other) const
		{
			return alpha == other.alpha && red == other.red && green == other.green && blue == other.blue;
		}
	};

	struct EXPORT_API ValueArea
	{
		int32_t width = 0;
		int32_t height = 0;

		bool operator==(ValueArea const& other) const
		{
			return width == other.width && height == other.height;
		}
	};

	EXPORT_API std::ostream& operator<<(std::ostream& os, ValueArea const& a);

	struct EXPORT_API ValueAreaFloat
	{
		double width = 0;
		double height = 0;

		bool operator==(ValueAreaFloat const& other) const
		{
			return std::abs(width - other.width) < 0.001 && std::abs(height - other.height) < 0.001;
		}
	};

	struct EXPORT_API ValuePoint
	{
		int32_t x = 0;
		int32_t y = 0;

		bool operator==(ValuePoint const& other) const
		{
			return x == other.x && y == other.y;
		}
	};

	EXPORT_API std::ostream &operator<<(std::ostream& os, ValuePoint const& p);

	struct EXPORT_API ValueRect
	{
		ValuePoint pos;
		ValueArea  size;

		bool operator==(ValueRect const& other) const
		{
			return pos == other.pos && size == other.size;
		}
	};

	EXPORT_API std::ostream &operator<<(std::ostream &os, ValueRect const& r);

	struct EXPORT_API ValueRange
	{
		int64_t min;
		int64_t max;

		bool operator==(ValueRange const& other) const
		{
			return min == other.min && max == other.max;
		}
	};

	struct EXPORT_API ValueRangeFloat
	{
		double min;
		double max;

		bool operator==(ValueRangeFloat const& other) const
		{
			return std::abs(min - other.min) < 0.001 && std::abs(max - other.max) < 0.001;
		}
	};


struct EXPORT_API CameraInfo
{
    bool mIsValid = false;
    TetherConnection mConnectionType = kTetherConnection_Undefined;
    MSVC_ALLOW_PUBLIC_STL(std::string mName);
    MSVC_ALLOW_PUBLIC_STL(std::string mIpV4);
    MSVC_ALLOW_PUBLIC_STL(std::string mSerial);

    static CameraInfo const Invalid;
};

class EXPORT_API ICameraCapabilities
{
public:
    /** @brief Holds the maximum number of pixels for a requested tile */
    virtual uint32_t MaxPixelsPerTile() const = 0;
};
EXPORT_API std::ostream& operator<<(std::ostream& os, const ICameraCapabilities& capabilities);

/**
 * @brief Base class used by all image browsing related classes
 *
 * Defines a common scheme to represent and query for error conditions.
 *
 * @since Version 3.0
 */
class EXPORT_API IImageFileCommon
{
public:
    virtual ~IImageFileCommon() {};

    enum class Errors : uint32_t
    {
        ErrorFlag_None                  = 0,
        ErrorFlag_Any                   = (1 << 0),
        ErrorFlag_QueueFull             = (1 << 1),
        ErrorFlag_UnsupportedFormat     = (1 << 2),
        ErrorFlag_FatalError            = (1 << 3),
        ErrorFlag_Unexpectedstatus      = (1 << 4),
        ErrorFlag_InvalidImageId        = (1 << 5),
        ErrorFlag_IsBusy                = (1 << 6),
        ErrorFlag_RequestFailed         = (1 << 7),
        ErrorFlag_InvalidBuffer         = (1 << 8),
        ErrorFlag_ImageSkipped          = (1 << 9),
        ErrorFlag_BufferRequestTimeout  = (1 << 10),
        ErrorFlag_FileIsCorrupt         = (1 << 11),
    };

    /** @brief The (session) unique image Id for this file. */
    virtual uint32_t    ImageId         () const = 0;

    /** @brief Returns whether the request resulted in any error(s). */
    virtual bool        AnyErrors       () const = 0;

    /** @brief The specific errors that occured. */
    virtual Errors      ImageFileErrors () const = 0;

    /** @brief A string version of the errors. */
    virtual std::string ErrorsString    () const = 0;
};


/**
 * @brief Interface that defines a basic set of properties for image files
 *        stored on the camera
 *
 * This class (interface) defines a curated set of common image metadata, that
 * can retrieved from files on the camera.
 *
 * @see P1::CameraSdk::Camera::BasicImageInfo
 * @since Version 3.0
 */
class EXPORT_API IImageFileBasicInfo : public virtual IImageFileCommon
{
public:

    /** @brief The possible white balance settings available in the camera */
    enum class WhiteBalanceSetting
    {
        Auto        = 0,
        Daylight    = 1,
        Fluorescent = 2,
        Tungsten    = 3,
        Flash       = 4,
        Custom1     = 5,
        Custom2     = 6,
        Custom3     = 7,
        Aerial      = 8
    };

    /** @brief Compression type that can be used in IIQ files */
    enum class CompressionSetting
    {
        None        = 0,
        IIQL        = 1,
        IIQS        = 2,
        IIQT        = 3,
        RAW16       = 4,
        IIQL16      = 5,
        IIQT16      = 6,
        RAW12       = 7,
        IIQL12      = 8,
        IIQT12      = 9
    };

    /** @brief Ways the image can by rotated, as defined by cameras orientation when capturing */
    enum class OrientationModeSetting : int32_t
    {
        Degrees0   = 0,
        Degrees90  = 90,
        Degrees180 = 180,
        Degrees270 = -90
    };

    /** @brief Users can assign ratings to image files, 1 to 5 stars */
    enum class StarRating
    {
        NotRated    = 0,
        OneStar     = 1,
        TwoStars    = 2,
        ThreeStars  = 3,
        FourStars   = 4,
        FiveStars   = 5
    };

    virtual ~IImageFileBasicInfo() {};

    /** @brief The Aperture of the capture, as Exif APEX value */
    virtual float ApertureApex() const = 0;

    /** @brief The ShutterSpeedValue of the capture, as Exif APEX value */
    virtual float ShutterSpeedApex() const = 0;

    /** @brief The horizantal (X) coordinate of the AFr focus point. 0.0 to 1.0, left to right. (-1.0 means invalid) */
    virtual float FocusPointX() const = 0;

    /** @brief The vertical (Y) coordinate of the AFr focus point. 0.0 to 1.0, top to bottom. (-1.0 means invalid) */
    virtual float FocusPointY() const = 0;

    /** @brief The focal length in mm of the lens used in capture */
    virtual float FocalLengthMm() const = 0;

    /** @brief The cameras white balance setting during the capture */
    virtual WhiteBalanceSetting WhiteBalance() const = 0;

    /** @brief The compression format used in the capture's IIQ file */
    virtual CompressionSetting IIQCompressionSetting() const = 0;

    /** @brief The cameras orientation (is image rotated) */
    virtual OrientationModeSetting Orientation() const = 0;

    /** @brief The valid width of the image in pixels. This is the width of the exposed sensor area. */
    virtual int32_t ImagePixelWidth() const = 0;

    /** @brief The valid height of the image in pixels. This is the height of the exposed sensor area. */
    virtual int32_t ImagePixelHeight() const = 0;

    /** @brief The ISO value of the capture. (Sensor sensitivity.) */
    virtual uint32_t Iso() const = 0;

    /** @brief Users rating of the image (1-5 stars) */
    virtual StarRating Rating() const = 0;

    /** @brief Get the time the capture was triggered, as a UNIX timestamp. (Seconds since epoch) */
    virtual uint32_t UnixTimeStamp() const = 0;

    /** @brief Get the total size of the IIQ image file , in bytes */
    virtual uint32_t ImageByteSize() const = 0;

    /** @brief Get the IIQ filename as it has been named on the cameras storages media */
    virtual std::string FileName() const = 0;

    /** @brief Get all the information formatted as a string */
    virtual std::string ToString() const = 0;
};

EXPORT_API std::ostream& operator<<(std::ostream& os, const IImageFileBasicInfo& basicImageInfo);


/**
 * @brief Contains an image buffer, encoded in either in bitmap or JPEG format.
 *
 * @since Version 3.0
 */
class EXPORT_API IImageBuffer
{
public:

    /** @brief The kind of image stored in the buffer */
    enum class Type
    {
        Unknown = 0,    /**< Undefined type */
        Preview = 1,    /**< The image is taken from the IIQ file's preview / thumbnail */
        Tile    = 2,    /**< The image is a tile used for tiling */
        LiveView= 3,    /**< Buffer is a LiveView frame */
        IIQ     = 4     /**< Buffer is an IIQ file */
    };

    virtual ~IImageBuffer() {};

    /** @brief Returns a pointer to the pixel buffer data */
    virtual std::shared_ptr<uint8_t const> Data() const = 0;

    /** @brief Returns the size of the pixel buffer in bytes */
    virtual size_t DataSizeBytes() const = 0;

    /** @brief Returns the image encoding type of this buffer (bitmap, jpeg) */
    virtual ImageFormat Format() const = 0;

    /** @brief Returns the type of this buffer */
    virtual Type BufferType() const = 0;

	/** @brief Returns a string of the type this buffer */
	virtual char const* BufferTypeStr() const = 0;

    /** @brief Returns a string of the type kind of this buffer, (RGB, JPEG, etc.) */
	virtual char const* FormatStr() const = 0;
};
EXPORT_API std::ostream& operator<<(std::ostream& os, IImageBuffer const& imageBuf);

/**
 * @brief Base class for those image buffers that specifies height and width.
 *
 * @since Version 3.0
 */
class EXPORT_API IImageBufferFrame : public virtual IImageBuffer
{
public:
    /** @brief Return the buffers width in pixels */
    virtual uint32_t DestinationWidth() const = 0;

    /** @brief Returns the buffers height in pixels */
    virtual uint32_t DestinationHeight() const = 0;
};
EXPORT_API std::ostream& operator<<(std::ostream& os, IImageBufferFrame const& imageBufFrame);

/**
 * @brief Contains a preview of an image, as returned from the camera.
 *
 * Either through subscriptions, or using *image browsing* you can get a preview
 * (aka. thumbnail) of an image.
 *
 * The camera normally creates a scaled RGB preview of the full image, right
 * after the capturing process. When such previews are send the SDK host, they
 * will come as implementations of this interface class.
 *
 * @since Version 3.0
 */
class EXPORT_API IPreview : public virtual IImageBufferFrame, public virtual IImageFileCommon
{};

/**
 * @brief A buffer that is a tile (crop) of an image.
 *
 * *Image browsing* enables zoom and pan, by delivering tiles of images using
 * this interface class.
 *
 * @since Version 3.0
 */
class EXPORT_API ITile : public virtual IImageBufferFrame, public virtual IImageFileCommon
{
public:

    /**
     * @brief This is a horizontal crop offset, in pixels.
     *
     * Origo is in upper left corner.
     */
    virtual uint32_t SourceOffsetX() const = 0;

    /**
     * @brief This is a vertical crop offset, in pixels.
     *
     * Origo is in upper left corner.
     */
    virtual uint32_t SourceOffsetY() const = 0;

    /**
     * @brief This is source (original) image width in pixels
     *
     * Tiling images can be scaled, such that source and destination dimentions differ. The source width
     * reopresents the original width of the crop of the source image on the camera.
     */
    virtual uint32_t SourceWidth() const = 0;

    /**
     * @brief This is source (original) image height in pixels
     *
     * Tiling images can be scaled, such that source and destination dimentions differ. The source height
     * reopresents the original height of the crop of the source image on the camera.
     */
    virtual uint32_t SourceHeight() const = 0;

    /** @brief This is the scaling factor that is applied to the tiled image. If any. */
    virtual float ScaleFactor() const = 0;
};
EXPORT_API std::ostream& operator<<(std::ostream& os, ITile const& imageBufFrame);

/**
 * @brief A buffer with a LiveView frame.
 *
 *
 *
 * @since Version 3.0
 */
class EXPORT_API ILiveView : public virtual IImageBufferFrame
{
public:

    /** @brief The compression quality as a percentage. */
    virtual uint32_t    CompressionQuality   () const = 0;

    /** @brief Returns whether the image has been compressed (e.g., by JPEG). */
    virtual bool        IsCompressed         () const = 0;

    /** @brief The actual crop that was cropped in the sensor. */
    virtual ValueRect   SourceCrop           () const = 0;

    /** @brief The user requested crop. */
    virtual ValueRect   RequestedCrop        () const = 0;

    /** @brief The full source (sensor) size. */
    virtual ValueArea   SourceFullImageSize  () const = 0;

};
EXPORT_API std::ostream& operator<<(std::ostream& os, ILiveView const& liveviewImage);

/**
 * @brief Contains a full IIQ image file.
 *
 * This means the entire file is *in memory*, and if written to a file, can be opened
 * by either Capture One or any other IIQ compatible image processing app.
 *
 * @since Version 3.0
 */
class EXPORT_API IFullImage : public virtual IImageBuffer
{
public:
    virtual std::string FileName () const = 0;
};
EXPORT_API std::ostream& operator<<(std::ostream& os, IFullImage const& fullImage);

/**
 * @brief Lightweight representation of an image file in the camera
 *
 * This class is the interface to image files returned by the camera. It is
 * lightweight and contains only a limited set of image metadata. It has no
 * actual image bitmap data.
 *
 * You use this *image item*'s *id* to query for extra info about an image or
 * the fetch preview or tiles of the image. You can also request the entire IIQ
 * file!
 *
 * @see P1::CameraSdk::Camera::AllImages
 * @since Version 3.0
 */
class EXPORT_API IStoredImageHandle
{
public:

    /**
     * @brief Bit mask that defines where an and how an image item is stored on
              camera
     *
     * Notice that multiple bits can be set, e.g. IIQ file has an accompanying
     * Jpeg or if image is present on _both_ XQD and SD.
     */
    enum class DeviceMask
    {
        None     = 0,
        Host     = 1, /**< IIQ transferred to (a) host in this (camera) power cycle */
        Xqd      = 2, /**< IIQ file of image is located on XQD card */
        Sd       = 4, /**< IIQ file of image is located on SD card */
        XqdJpeg  = 8, /**< Jpeg file of image is located on XQD card */
        SdJpeg   = 16 /**< Jpeg file of image is located on SD card */
    };

    virtual ~IStoredImageHandle() {};

    /** @brief Gets the *id* of this image. Can change after a camera reboot or a storage swap. */
    virtual uint32_t Id() const = 0;

    /**
     * @brief Gets the capture date & time for this image. As a Unix Epoch
     *        timestamp. (Seconds since 1970-01-01 00:00:00 UTC.)
     */
    virtual uint32_t Date() const = 0;

    /** @brief Gets the filename of the image's IIQ file */
    virtual std::string const& Filename() const = 0;

    /**
     * @brief Gets a bitmask defining on what devices (XQS/SD), the image is
     *        stored.
     */
    virtual DeviceMask Device() const = 0;

    /**  @brief Device() in a string format */
    virtual std::string DeviceString() const = 0;

    /**
     * @brief Requests a set of commonly used meta data about the image.
     *
     * CameraSDK defines a curated set of common meta data about an image
     * called *BasicImageInfo*. See @ref IImageFileBasicInfo for what meta
     * data is is included.
     *
     * The requested *basic info* object is returned by a notification. To
     * retrieve the object, you must listen for the
     * @ref EventType::CameraImageBasicImageInfo notification, _before_
     * calling this method.
     *
     * @see BasicImageInfo
     * @since Version 3.0
     */
    virtual void RequestBasicImageInfoAsync() const = 0;

    /**
     * @brief A blocking version of @ref RequestBasicImageInfoAsync.
     *
     * @param timeoutMs An optional maximum amount of milliseconds to wait
     *                  for a _basic info_ object, before giving up.
     *                  Zero (0) is *wait forever*.
     * @return A pointer to an @ref IImageFileBasicInfo object or nullptr
     *         on timeout.
     * @see RequestBasicImageInfoAsync
     * @since Version 3.0
     */
    virtual std::shared_ptr<IImageFileBasicInfo const> BasicImageInfo(uint32_t const timeoutMs = 0) const = 0;

    /**
     * @brief Request a full IIQ file of the image
     *
     * This method triggers on-demand transfer of the entire IIQ file of an
     * image. To retrieve the object, you must listen for the
     * @ref EventType::CameraImageReady notification, _before_ calling this method.
     * This notification contains the image via the FullImage() method.
     *
     * The listener thread cannot know whether the arrived image is a just triggered capture,
     * or a result of an on-demand request from this method.
     * Therefore, use the IIQ filename to identify your requested image.
     *
     * @note It is necessary to subscribe to image transfers via IFullImageSubscription
     *        for images to be transferred.
     *
     * @see IFullImageSubscription
     * @since Version 3.0
     */
    virtual bool RequestImageAsync() const = 0;

    /**
     * @brief Request the preview (thumbnail) for an image.
     *
     * A preview is a scaled down version of the entire image, encoded as either a RGB bitmap or as
     * an complete jpeg file. Depending on the requested `fileFormat` format, the camera might be able
     * to use any existing preview (thumbnail) present inside the image's IIQ file.
     *
     * The RGB previews can be stored inside the IIQ files. This method returns the stored preview, if it
     * exists, or generates a new one if not. If you ask for a Jpeg formatted preview, the generated or
     * stored RGB preview is converted to Jpeg (on the camera), before being returned to you. Jpeg
     * previews are not cached.
     *
     * Jpegs are significantly smaller in file size, but time consuming to encode. Prefer Jpeg if you are
     * connected to the camera over a slow connection, like an outdoor Wifi.
     *
     * The dimensions of the preview is controlled by the camera.
     *
     * The requested preview image object is returned by a notification. To
     * retrieve the object, you must listen for the
     * @ref EventType::CameraPreviewReady notification, _before_
     * calling this method.
     *
     * @param fileFormat Optionally change encoding of the returned preview image. Default is RGB bitmap.
     * @see RequestTileAsync
     * @since Version 3.0
     */
    virtual void RequestPreviewAsync(ImageFormat const fileFormat = ImageFormat::RGB24) const = 0;

    /**
     * @brief A blocking version of @ref RequestPreviewAsync.
     *
     * @param fileFormat Optionally change encoding of the returned preview image. Default is RGB bitmap.
     * @param timeoutMs An optional maximum amount of milliseconds to wait
     *                  for the preview, before giving up. Zero (0) is
     *                  'wait forever'.
     *
     * @return A pointer to an image buffer that encapsulates the image data, or nullptr on timeout.
     * @since Version 3.0
     */
    virtual std::shared_ptr<IPreview const> Preview(
        ImageFormat const fileFormat = ImageFormat::RGB24,
        uint32_t const timeoutMs = 0) const = 0;

    /**
     * @brief Input parameter for requesting a tile.
     *
     * @param x The horizontal pixel offset of the source crop
     * @param y The vertical pixel offset of the source crop
     * @param src Width The pixel width of the source crop
     * @param src Height The pixel height of the source crop
     * @param scaleFactor Optionally apply any down-scaling of the destination image tile: 0.0 < scaleFactor <= 1.0.
     * @param fileFormat Optionally define another image encoding format, than RGB24
     */
    struct TileRequest
    {
        uint32_t x;
        uint32_t y;
        uint32_t srcWidth;
        uint32_t srcHeight;
        float    scaleFactor = 1.0;
        ImageFormat fileFormat = ImageFormat::JPEG;
    };

    /**
     * @brief Request a _tile_: a (scaled) crop of the full image
     *
     * Tiles are a cropped and scaled jpeg encoded previews of the full image. The camera takes the
     * requested crop of the image, scales it and, optionally, converts it to jpeg.
     *
     * The crop is defined in full image coordinates, meaning the pixel range of the full (visible) image
     * dimensions. Origo is upper left corner. Any scaling is applied _after_ cropping, meaning that if
     * the crop dimension is 100 px, scaling by 0.5 results in an output image that is ~50 px.
     *
     * This method is asynchronous and will return immediately. The requested tile object is returned
     * by a notification. To retrieve the object, you must listen for the
     * @ref EventType::CameraTileReady notification, _before_ calling this method. Note that this
     * notification @ref EventType is shared with incoming *previews*.
     *
     * Note that the camera might not (and will likely not), give the exact crop you requested. The
     * resulting might differ a few pixels in offset or size. This is due performance optimizations in the
     * image conversion pipeline. Speed is favoured over precision.
     *
     * @param tileRequest Specification of requested tile.
     * @see RequestPreviewAsync
     * @since Version 3.0
     */
    virtual void RequestTileAsync(TileRequest const& tileRequest) const = 0;

    /**
     * @brief A blocking version of @ref RequestTileAsync.
     *
     * @param tileRequest Specification of requested tile.
     * @param timeoutMs To maximum time to wait for the camera to return the
     *                  tile. Zero (0) is 'wait forever'.

     * @return A pointer to an image buffer that encapsulates the image data, or nullptr on timeout.
     * @see Preview
     * @since Version 3.0
     */
    virtual std::shared_ptr<ITile const> Tile(
        TileRequest const& tileRequest,
        uint32_t timeoutMs = 0
    ) const = 0;

    /**
     * @brief Delete an image from the camera storage.
     *
     * This method deletes the image (IIQ file) and any associated Jpeg file,
     * from either XQD or SD card storage. Deletion is irreversible.
     *
     * Confirmation of delete request is provided by a notification. If you wish to receive this notification,
     * you must listen for the @ref EventType::CameraImageEnumElementRemoved notification,
     * _before_ calling this method. If the request fails, there is no notification.
     *
     * @since Version 3.0
     */
    virtual void DeleteImage() = 0;

    /**
     * @brief Change the star-rating of an image.
     *
     * All images can have an optional star-rating, to mark them as more or less significant. Use this
     * method to change an image rating, or remove it.
     *
     * Confirmation of a changed rating is provided by a notification. If you wish to receive this
     * notification, you must listen for the @ref EventType::CameraImageEnumElementEdited
     * notification, _before_ calling this method. If the request fails, there is no notification.
     *
     * To retrieve the existing star-rating of an image you should use the methods for retrieving all
     * images on the camera, or the @ref BasicImageInfo.
     *
     * @param rating The new rating to set on the image
     * @since Version 3.0
     */
    virtual void RateImage(IImageFileBasicInfo::StarRating const rating) = 0;
};

/**
 * @brief v
 *
 * Used by CameraSDK internals to decide how (and if) an incoming
 * @ref INotificationEvent should be added to @ref Listener queues.
 *
 * Each *notification event* defines a QueueingStrategy, that is consulted when
 * new notifications are added to the @ref Listener.
 *
 * @since Version 3.0
 */
enum class QueueingStrategy
{
    Always          = 0,/**< Just enqueue the event, no matter what */
    Overwite        = 1,/**< Overwrite a similar already existing event on the queue */
    DiscardIfExists = 2 /**< If a similar event is already present, discard / ignore new ones */
};

/**
 * @brief Notification interface that encapsulate all types of notifications
 *
 * Notifications relate to specific events. These events might come with data
 * attached. This notification interface, handles all types of data events can
 * provide.
 *
 * Events have a *type* and a *category*. Multiple types shares the same
 * category. Like the types: `CameraImageReady` and `CameraLiveViewStarted` are
 * both in the `Camera` category. The category says something about what
 * auxiliary data is provided with the notification.
 *
 * You can use the `Is**Event()` helper methods to query what kind of category
 * a notification event belongs to.
 *
 * @since Version 3.0
 */
class EXPORT_API INotificationEvent
{
public:

    
    virtual ~INotificationEvent() {};

    virtual bool operator==(INotificationEvent const& other) const = 0;
    virtual bool operator!=(INotificationEvent const& other) const = 0;

    /** @brief Get a reference the EventType this notification comes from */
    virtual BaseEventType const& Type() const = 0;

    /** @brief `true` if this notification event has *filename* data attached */
    virtual bool HasImageFilename() const = 0;

    /**
     * @brief Gets the *image id* of the image this event relates to, or zero
     *        (0) if this is not an image related event.
     */
    virtual uint32_t ImageId() const = 0;

    /** @brief `true` if this notification event relates to specific camera */
    virtual bool HasCamId() const = 0;

    /**
     * @brief Gets the *id* of the camera that the notification relates to.
     */
    virtual CameraId CamId() const = 0;

    /**
     * @brief If this is a property related event (category is 'Property'), then
     *        this is the *id* of that property.
     */
    virtual uint32_t PropertyId() const = 0;

    /** @brief `true` if this notification event is of category *Camera* */
    virtual bool IsCameraEvent() const = 0;

    /** @brief `true` if this notification event is of category *Sdk* */
    virtual bool IsSdkEvent() const = 0;

    /** @brief `true` if this notification event is of category *Property* */
    virtual bool IsPropertyEvent() const = 0;

    /** @brief `true` if this notification event is of category *Transfer* */
    virtual bool IsTransferEvent() const = 0;

    /** @brief `true` if this notification event is of category *Image* */
    virtual bool IsImageEvent() const = 0;

    /**
     * @brief If event type is @ref CameraImageBasicImageInfo, this holds the
     *        basic info data object.
     */
    virtual std::shared_ptr<IImageFileBasicInfo const> BasicImageInfo() const = 0;

    /**
     * @brief If event type is @ref CameraPreviewReady, this holds the image
     *        buffer data object.
     */
    virtual std::shared_ptr<IPreview const> Preview() const = 0;

    /**
     * @brief If event type is @ref CameraTileReady, this holds the image
     *        buffer data object.
     */
    virtual std::shared_ptr<ITile const> Tile() const = 0;

    /**
     * @brief If event type is @ref CameraNewLiveViewImage, this holds the image
     *        buffer data object.
     */
    virtual std::shared_ptr<ILiveView const> LiveView() const = 0;

    /**
     * @brief If event type is @ref CameraImageReady, this holds the full image
     */
    virtual std::shared_ptr<IFullImage const> FullImage() const = 0;

    /**
     * @brief If event type is @ref CameraImageAdded, @ref CameraImageRemoved
     *        or @ref CameraImageEdited this holds the image list item object.
     */
    virtual std::shared_ptr<IStoredImageHandle> StoredImageHandle() const = 0;

    /**
     * @brief If event type is of *Sdk* category, this holds basic camera info
     *        data object.
     */
    virtual CameraInfo const& CamInfo() const = 0;

    /**
     * @brief If event type is @ref ImageTransferProgress, this returns the
     *        percent progress.
     */
    virtual uint8_t TransferProgressPercent() const = 0;

    /**
     * @brief If event type is of *Transfer* category, this contains the
     *        image file's filename.
     */
    virtual std::string ImageFilename() const = 0;

    /**
     * @brief Returns a short (human readable) description of the notification
     *        event
     *
     * This string description include the event type and any auxiliary data
     */
    virtual std::string ToString() const = 0;

    /**
     * @brief Returns how this notification is queued in @ref Listener
     *
     * Each *NotificationEvent* implementation defines a strategy of how it is
     * added to notification queues.
     *
     * @sa QueueMode
     */
    virtual QueueingStrategy QueueMode() const = 0;
};

using NotificationEventPtr = std::shared_ptr<INotificationEvent const>;

#pragma pack(pop)

EXPORT_API std::ostream& operator<<(std::ostream& os, Version const& version);

}  // CameraSdk
}  // P1

#endif // C_P1CAMERA_COMMON_STRUCTS_H
